/* ple.c */

#ifdef __APPLE__
#include <OpenGL/gl.h>
#else
#include <GL/glew.h>
#include <GL/gl.h>
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "plot.h"

static Scene *scn = NULL;
static Matrix4 nm, mv, mvp, mdpy;
static void *vdata;


#define SHADE_NPOLYS 1, 3
#define PL(l) (l)
#define TAIL(l) (l->next)

// -------------------------------------------------

Hpoly *prim_polys(Prim *s, iTri *t);
Hpoly *mesh_polys(iMesh *m, iTri *t);

// -------------------------------------------------


int render_scene(Scene *s)
{
    Object *o;
    iMesh *im; iTri *t;
    Hpoly *pl;
 
    if (s == NULL) return 0;

    scn = s;
    init_render(s);
    s->objs = graph_flatten(graph_transform(s->objs));

    for (o = s->objs; o != NULL; o = o->next) {
        im = tesselate(o);
        for (t = im->t; t != NULL; t = t->next) {
            pl = assembly(o, t);
            if (!vertex_proc(pl, vdata, o))
                continue;
            if ((poly_clip(pl, efree, 0)) != 0) {
                hpoly_transform(hpoly_homo(pl), mdpy);
                plot_hpoly(pl);
            }
	    hplist_free(pl);
        }
    }
    close_render();  
    return TRUE;
}


/* ------------------ stage ops  -------------------- */


void init_render(Scene *s)
{
  mv = s->view->V; nm = s->view->Vinv;
  mvp =  m4_m4prod(m4_m4prod(s->view->P, s->view->C), s->view->V);
  mdpy = s->view->S;

  dvwindow(s->img->w, s->img->h);
  dvclear(0.0);
  dvcolor(1.0);
}

iMesh *tesselate(Object *obj)
{
  switch (obj->type) {
  case V_PRIM:  return prim_uv_decomp(prim_(obj), 1.);
  case V_TRIMESH: return mesh_(obj);
  }
  return NULL;
}

Hpoly *assembly(Object *obj, iTri *t)
{
  switch (obj->type) {
  case V_PRIM: return prim_polys(prim_(obj), t);
  case V_TRIMESH: return mesh_polys(mesh_(obj), t);
  }
  return NULL;
}

int vertex_proc(Hpoly *p, void *d, Object *obj)
{
  if (is_backfacing(p, v3_sub(hpoly_centr(p), scn->view->center)))
    return FALSE;
  hpoly_transform(p, mvp);
  return TRUE;
}

void plot_hpoly(Hpoly *p)
{
  int i;
  dvcolor(1.0);

  for (i = 0; i < p->n; i++) {
    int i1 = (i == p->n-1 )? 0 : i+1;
    dvline(p->v[i].x, p->v[i].y, p->v[i1].x, p->v[i1].y);
  }

}

void close_render(void)
{
  dvflush();
}

/* -------------------- obj dependent ----------------------- */

Hpoly *prim_polys(Prim *s, iTri *t)
{
  int i;
  Hpoly *pl = hplist_alloc(SHADE_NPOLYS);
  for (i = 0; i < 3; i++) {
    Vector3 p = imesh_mpoint(t->m, t->v[i]);
    PL(pl)->v[i] = v4_v3conv(prim_point(s, p.x, p.y));
  }
  TAIL(pl) = NULL;
  return pl;
}

Hpoly *mesh_polys(iMesh *m, iTri *t)
{
  int i;
  Hpoly *pl = hplist_alloc(SHADE_NPOLYS);
  for (i = 0; i < 3; i++) {
    PL(pl)->v[i] = v4_v3conv(imesh_point(m, t->v[i]));
  }
  TAIL(pl) = NULL;
  return pl;
}


// -------------------- DV ---------------------


static double dv_w, dv_h;

void dvflush(void)
{
    glFlush();
}

void dvclear(double c)
{
    glClearColor(c, c, c, 0.f);
    glClear(GL_COLOR_BUFFER_BIT);
}

void dvcolor(double c)
{
    glColor3d(c, c, c);
}

void dvwindow(double xmax, double ymax)
{
    dv_w = xmax;
    dv_h = ymax;
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
    glOrtho((GLdouble)(0), (GLdouble)(xmax), (GLdouble)(0), (GLdouble)(ymax), -1.0, 1.0);
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity();
}

void dvplot(double x1, double y1)
{
    double s = 5.0;
    glRectd((GLdouble)(x1-s), (GLdouble)(y1-s), (GLdouble)(x1+s), (GLdouble)(y1+s));
}

void dvline(double x1, double y1, double x2, double y2)
{
    glBegin(GL_LINES);
    glVertex2d((GLdouble)(x1), (GLdouble)(y1));
    glVertex2d((GLdouble)(x2), (GLdouble)(y2));
    glEnd();
}

double dv_map_y(double y)
{
    return dv_h - y;
}
