#include "main.h"

#define RD_EPS   2
#define MAX_ITER 100
//#define VISIB    FALSE
#define VISIB    TRUE
#define ELEM_SIZ 80

static Scene *s;
static RadData *rd;
static Matrix4 mclip, mdpy;


int main(int argc, char **argv)
{
  Hpoly *p; Color c, *cols; int k;
  Hpoly *h = hpoly_alloc(3);

  init_sdl();
  s = scene_read();
  init_render();

  radiosity_set(RD_EPS, MAX_ITER, VISIB);
  cols = radiosity_prog(rd->n, rd->pl, rd->e, rd->rho);

  for (k = 0; k < rd->n; k++) {
    p = rd->pl[k];
    h = hpoly3_set(h, rd->pl[k]->v[0], rd->pl[k]->v[1], rd->pl[k]->v[2]); 
    if (is_backfacing(h, v3_sub(hpoly_centr(h), s->view->center)))
      continue;
    c = cols[k];
    hpoly_transform(h, mclip);
    if (poly_clip(h, efree, 0)) {
      hpoly_transform(hpoly_homo(h), mdpy);
      scan_hpoly(h, pix_paint, &c);
    }
  }
  img_write(s->img, "stdout", 0);
  efree(h->v); efree(h);
  exit(0);
}


void pix_paint(Vector4 v,int n,int lv,Real lt,int rv,Real rt,Real st, void *c)
{
  if (zbuf_store(v))
    img_putc(s->img, v.x, v.y, col_dpymap(*((Color *)(c))));
}

void init_render(void)
{
  mclip = m4_m4prod(s->view->P, m4_m4prod(s->view->C, s->view->V));
  mdpy = s->view->S;

  zbuf_init(s->img->w, s->img->h);
  rd = make_rad_data(s->objs);
} 

void init_sdl(void)
{
  lang_defun("scene", scene_parse);
  lang_defun("view", view_parse);
  lang_defun("dist_light", distlight_parse);
  lang_defun("plastic", plastic_parse);
  lang_defun("polyobj", obj_parse);
  lang_defun("trilist", htrilist_parse);

  lang_defun("include", include_parse);
}

RadData *make_rad_data(Object *ol)
{
  Object *o; Hpoly *p; int k;
  RadData *rd = NEWSTRUCT(RadData);

  for (rd->n = 0, o = ol; o != NULL; o = o->next) {
    if (o->type == V_HPOLYLIST) {
      p = o->u.sh->u.hpols;
      o->u.sh->u.hpols = hplist_refine(p, ELEM_SIZ, 4);
      rd->n += (k = hplist_lenght(o->u.sh->u.hpols));
      hplist_free(p);
    }
  }
  rd->pl = NEWARRAY(rd->n, Hpoly *);
  rd->e = NEWARRAY(rd->n, Vector3);
  rd->rho = NEWARRAY(rd->n, Vector3);

  for (k = 0, o = ol; o != NULL; o = o->next) {
    if (o->type != V_HPOLYLIST)
      continue;
    for (p = o->u.sh->u.hpols; p != NULL; p = p->next) {
      rd->pl[k] = p;
      rd->rho[k] = c_scale(o->mat->kd, o->mat->c);
      rd->e[k] = c_scale(o->mat->ks, o->mat->s);
      k++;
    }
  }
  return rd;      
}
