#include "main.h"

static Scene *s;
static Object *o;
static Matrix4 mclip, mdpy;
static RContext *rc;
static void *vdata;
static List *z = NULL;


int main(int argc, char **argv)
{
  iMesh *im; iTri *t;
  Hpoly *p; Item *i;

  init_lang();
  s = scene_read();
  s->objs = graph_flatten(graph_transform(s->objs));
  init_render();
  for (o = s->objs; o != NULL; o = o->next) {
    im = tesselate(o);
    for (t = im->t; t != NULL; t = t->next) {
      p = assembly(o, t);
      if (!vertex_proc_cull(p, vdata, o))
	continue;
      hither_clip(0, p, z_store, hplist_free);
    }
  }
  z = z_sort(z);
  for (i = z->head; i != NULL; i = i->next) {
    p = vertex_proc_shade(p, vdata, i);
    scan_hpoly(p, pixel_proc, vdata);
    hplist_free(p);
  }
  raster_proc(s);
  exit(0);
}

void z_store(Hpoly *l)
{
  Zdatum *d = NEWSTRUCT(Zdatum);

  d->zmax = MIN(l->v[0].z, MIN(l->v[1].z, l->v[2].z));
  d->l = l;
  d->o = o;
  append_item(z, new_item(d));
}

/* ------------------ stage ops  -------------------- */

void init_render(void)
{
  mclip = m4_m4prod(s->view->C, s->view->V);
  mdpy = m4_m4prod(s->view->S, s->view->P);
  z = new_list();
  vdata = NEWSTRUCT(SHADE_DATA);
  rc_sset(rc = NEWSTRUCT(RContext), s->view, s->lights, s->img);
}

iMesh *tesselate(Object *obj)
{
  switch (obj->type) {
  case V_PRIM:  return prim_uv_decomp(prim_(obj), 1.);
  case V_TRIMESH: return mesh_(obj);
  }
  return NULL;
}

Hpoly *assembly(Object *obj, iTri *t)
{
  switch (obj->type) {
  case V_PRIM: return prim_polys(prim_(obj), t);
  case V_TRIMESH: return mesh_polys(mesh_(obj), t);
  }
  return NULL;
}

int vertex_proc_cull(Hpoly *p, void *d, Object *obj)
{
  if (is_backfacing(p, v3_sub(hpoly_centr(p), s->view->center)))
    return FALSE;
  hpoly_homoxform(p, mclip);
  return TRUE;
}

Hpoly *vertex_proc_shade(Hpoly *p, void *d, Item *i)
{
    SHADE_PREP(SHADE_PREP_PARAMS);
    p = hpoly_homoxform(P(i), mdpy);
    vdata = SHADE_SET(SHADE_SET_PARAMS);
    return p;
}

void pixel_proc(Vector4 v,int n,int lv,Real lt,int rv,Real rt,Real st,void *c)
{
    SHADE_PAINT(v, n, lv, lt, rv, rt, st, c);
}

void raster_proc(Scene *s)
{
  img_write(s->img, "stdout", 0);
}

/* ------------------- obj dep ------------------- */

Hpoly *prim_polys(Prim *s, iTri *t)
{
  int i;
  Hpoly *l = hplist_alloc(SHADE_NPOLYS);

  for (i = 0; i < 3; i++) {
    Vector3 p = imesh_mpoint(t->m, t->v[i]);
    PL(l)->v[i] = v4_v3conv(prim_point(s, p.x, p.y));
#if defined(GOURAUD_SHADE) 
    NL(l)->v[i] = v4_v3conv(prim_normal(s, p.x, p.y));
#endif
#ifdef GOURAUD_SHADE
    CL(l)->v[i] = v4_make(0,0,0,0);
#endif
  }
  TAIL(l) = NULL;			  
  return l;
}

Hpoly *mesh_polys(iMesh *m, iTri *t)
{
  int i;
  Hpoly *l = hplist_alloc(SHADE_NPOLYS);

  for (i = 0; i < 3; i++) {
    PL(l)->v[i] = v4_v3conv(imesh_point(m, t->v[i]));
#if defined(GOURAUD_SHADE)
    NL(l)->v[i] = v4_v3conv(imesh_normal(m, t->v[i]));
#endif
#ifdef GOURAUD_SHADE
    CL(l)->v[i] = v4_make(0,0,0,0);
#endif
  }
  TAIL(l) = NULL;
  return l;
}

void init_lang(void)
{
  lang_defun("scene", scene_parse);
  lang_defun("view", view_parse);
  lang_defun("dist_light", distlight_parse);
  lang_defun("plastic", plastic_parse);
  lang_defun("primobj", obj_parse);
  lang_defun("sphere", sphere_parse);
  lang_defun("meshobj", obj_parse);
  lang_defun("mesh", imesh_parse);
  lang_defun("group", group_parse);
  lang_defun("affine", xaffine_parse);

  lang_defun("include", include_parse);
}



